% 
% simulation relative to theorem 2 
% of  the paper
%   Flatness-based control of a single qubit gate
% 
%   Paulo Sergio Pereira da Silva 
%   Pierre Rouchon
% 
%    December 2006
clear all 
%
T=2;   % steering time from 0 to T
ds=1.e-4; % sampling step in s-scale [0,1]
s=[-2*ds:ds:1+2*ds];
dt=T*ds; t=T*s;
% time scaling function s=zeta(t);
zeta=3*(t/T).^2-2*(t/T).^3;
%nu=2.5;zeta=(abs(t/T)).^nu./((abs(t/T)).^nu+(1-(abs(t/T))).^nu); 
%
% numeric derivation  versus s and t
N=length(s)-4;II=[3:1:length(s)-2];
ii=[1:1:length(s)];iim1=max(1,ii-1);iip1=min(length(s),ii+1); 
%
% random goal quaternion 
 qb0=1-2*rand; qb1=(1-2*rand); qb2=(1-2*rand); qb3=(1-2*rand); 
 Ld=(qb0^2+qb1^2+qb2^2+qb3^2)^(0.5);
 qb0=qb0/Ld;qb1=qb1/Ld;qb2=qb2/Ld;qb3=qb3/Ld;
qb0=0;qb1=0;qb2=0;qb3=1; % Goal quaternion of the paper
%
%  rotation of etab for control
  etab=pi/2-angle(qb1+sqrt(-1)*qb2); 
% 
% goal quaternion after rotation of angle etab 
  yb0=qb0; yb1=0; yb2=sqrt(qb1^2+qb2^2);yb3=qb3;
  betab=angle(yb2+sqrt(-1)*yb3);
  alphab=angle(yb0+sqrt(-1)*(cos(betab)*yb2+sin(betab)*yb3)); 
% right translation via 
%   cos(lambdab) + sin(lambdab)*(cos(betab)*e2+sin(betab)*e3)
if  ( (abs(alphab) < pi/4) )
    lambdab=pi/4-alphab/2;
elseif (abs(alphab) < 3*pi/4)
    lambdab=-alphab/2; 
else
    lambdab=pi/4-alphab/2;
end
%
% initial and final constraints on alpha et beta in s scale
alpha0=lambdab;
alpha1=lambdab+alphab; 
beta0=betab;
beta1=betab;
dalpha0=cos(beta0)*alphab;
dalpha1=cos(beta1)*alphab;
dbeta0=-sin(beta0)/(sin(alpha0)*cos(alpha0)) * alphab;
dbeta1=-sin(beta1)/(sin(alpha1)*cos(alpha1)) * alphab;
% 
% polynomial of degree 3 in s for alpha and beta  
C=3*(alpha1-alpha0-dalpha0)-(dalpha1-dalpha0);
D=(dalpha1-dalpha0)- 2*(alpha1-alpha0-dalpha0);
alpha= alpha0+dalpha0*s+C*s.^2+D*s.^3;
C=3*(beta1-beta0-dbeta0)-(dbeta1-dbeta0);
D=(dbeta1-dbeta0)- 2*(beta1-beta0-dbeta0);
beta= beta0+dbeta0*s+C*s.^2+D*s.^3;
%
% the flat output 
y0=cos(alpha);y1=0*y0;y2=sin(alpha).*cos(beta);y3=sin(alpha).*sin(beta);
% numerical derivation versus s
Dy0=(y0(iip1)-y0(iim1))/(2*ds);
Dy1=(y1(iip1)-y1(iim1))/(2*ds);
Dy2=(y2(iip1)-y2(iim1))/(2*ds);
Dy3=(y3(iip1)-y3(iim1))/(2*ds);
%
w1= y0.*Dy1-Dy0.*y1 + y2.*Dy3-Dy2.*y3;
w2= y0.*Dy2-Dy0.*y2 + y3.*Dy1-Dy3.*y1;
w3= y0.*Dy3-Dy0.*y3 + y1.*Dy2-Dy1.*y2;
%
Dw2=(w2(iip1)-w2(iim1))/(2*ds);
Dw3=(w3(iip1)-w3(iim1))/(2*ds);
%
Dzeta=(zeta(iip1)-zeta(iim1))/(2*dt);
%
% remove  the 2 initial and 2 final values 
t=t(II);s=s(II);zeta=zeta(II);Dzeta=Dzeta(II);
alpha=alpha(II);beta=beta(II);
y0=y0(II);y1=y1(II);y2=y2(II);y3=y3(II);
w1=w1(II);w2=w2(II);w3=w3(II);Dw2=Dw2(II);Dw3=Dw3(II);
% synchro between s and zeta
Nzeta=round(zeta/ds)+1;
%
z=w2-sqrt(-1)*w3; phi= angle(z)/2;
%
u1b=w1+(Dw2.*w3-Dw3.*w2)./(2*(w2.^2+w3.^2));
u2b=(w2.^2+w3.^2).^(0.5);
% control in the s-scale
u1s= cos(etab)*u1b + sin(etab)*u2b;
u2s= -sin(etab)*u1b + cos(etab)*u2b;
% control in the t-scale
u1=Dzeta.*u1s(Nzeta);
u2=Dzeta.*u2s(Nzeta);
% verification  via numerical simulation 
% that with  u1 and u2 one has q(T)=qb
% when q(0)=1
q0=zeros(1,N);q1=q0;q2=q0;q3=q0; q0(1)=1;
for n=2:N;
    Dq0= -u1(n-1)*q1(n-1)-u2(n-1)*q2(n-1);
    Dq1=  u1(n-1)*q0(n-1)+u2(n-1)*q3(n-1);
    Dq2= -u1(n-1)*q3(n-1)+u2(n-1)*q0(n-1);
    Dq3=  u1(n-1)*q2(n-1)-u2(n-1)*q1(n-1);

    q0(n)=q0(n-1)+dt*Dq0;
    q1(n)=q1(n-1)+dt*Dq1;
    q2(n)=q2(n-1)+dt*Dq2;
    q3(n)=q3(n-1)+dt*Dq3;
    
    Lq=(q0(n)^2+q1(n)^2+q2(n)^2+q3(n)^2)^(0.5);
    q0(n)=q0(n)/Lq;
    q1(n)=q1(n)/Lq;
    q2(n)=q2(n)/Lq;
    q3(n)=q3(n)/Lq;
end

figure(1), clf;  % figure of the paper with qb=e3 
   subplot(211)
     plot(t,q0,'k-',t,q1,'r-',t,q2,'b-',t,q3,'m-',T,qb0,'ks',T,qb1,'rs',T,qb2,'bs',T,qb3,'ms','linewidth',2);
     legend('q_0','q_1','q_2','q_3');
     xlabel('time')
   subplot(212)
     plot(t,u1,'r-',t,u2,'b--','linewidth',2);
     legend('u_1','u_2');
     xlabel('time')
     
figure(2),clf;
subplot(221);
  plot(s,y0,s,y1,s,y2,s,y3);
  legend('y_0','y_1','y_2','y_3');
  xlabel('s')
subplot(222);
  plot(s,w1,s,w2,s,w3);
  legend('w_1','w_2','w_3');
  xlabel('s')
subplot(223);
  plot(s,phi);
  legend('\phi');
  xlabel('s')
subplot(224);
  plot(s,u1s,s,u2s);
  legend('u_1','u_2');
    xlabel('s')
    
figure(3),clf;
  plot(s,alpha,s,beta);
  legend('alpha','beta');
  xlabel('s')

figure(1);

return
  
  
  



